/** ***********************************************************************
*
* Veracity Technology Consultants CONFIDENTIAL
* __________________
*
*  2019 Veracity Technology Consultants
*  All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains
* the property of Veracity Technology Consultants and its suppliers,
* if any.  The intellectual and technical concepts contained
* herein are proprietary to Veracity Technology Consultants
* and its suppliers and may be covered by U.S. and Foreign Patents,
* patents in process, and are protected by trade secret or copyright law.
* Dissemination of this information or reproduction of this material
* is strictly forbidden unless prior written permission is obtained
* from Veracity Technology Consultants.
*/
const LrsPlugin = require('./utils/plugins/lrsPlugin');
const utils = require('./utils/utils');
const express = require('express');
const moment = require('moment');
const QueryStage = require('./analytics/engine/stages/QueryStage');
const Executor = require('./analytics/engine/Executor');
const { getUniqueActivityDocs } = require('./utils');

module.exports = class DcsaReports extends LrsPlugin {
    constructor(lrs, dal, settings) {
        super(lrs, dal, settings);

        this.on('lrsHomeLinks', (event, lrs) => ({
            text: 'DCSA Reports',
            href: this.getLink('/dcsaReports'),
            subtext: 'Generate and download course reports.',
        }));

        const router = express.Router();
        router.use(utils.ensureLoggedIn);

        const padTo2Digits = (num) => num.toString().padStart(2, '0');

        const deliveryMethods = [
            { value: "instructor", display: "Instructor" },
            { value: "technology", display: "Technology" },
        ];

        const reports = [
            { value: "details", display: "Details" },
            { value: "militaryStatusActivityType", display: "Military Status By Activity Type" },
            { value: "disciplineMilitaryStatus", display: "Military Status By Discipline" },
            { value: "organizationMilitaryStatus", display: "Military Status By Organization" },
            { value: "activityTypeOrganization", display: "Organization By Activity Type" },
            { value: "courseCompletion", display: "Course Completion By Division" },
        ];

        const courseStatus = [
            { value: "all", display: "All" },
            { value: "completed", display: "Completed" },
            { value: "notCompleted", display: "Not Completed" },
        ];

        const deliveryMethodFilter = (method) => (
            method === 'instructor' ? [
                'Instructor-led', 'Virtual Instructor-led'
            ] : [
                'Job Aid',
                'eLearning',
                'Short',
                'Curriculum'
            ]
        );

        const getHomeParams = async (error, level) => {
            const dateQuery = {
                filter: {
                    "object.definition.type": "http://adlnet.gov/expapi/activities/course",
                },
                process: [
                    {
                        $parallel: {
                            stages: [
                                {
                                    $query: {
                                        filter: {

                                        },
                                        process: [
                                            {
                                                $min: { path: 'timestamp' }
                                            },
                                        ]
                                    }
                                },
                                {
                                    $query: {
                                        filter: {

                                        },
                                        process: [
                                            {
                                                $max: { path: 'timestamp' }
                                            }
                                        ]
                                    }
                                },
                            ]
                        }
                    },
                    {
                        $mergeArrays: {},
                    }
                ]
            };

            const queryStage = new QueryStage({
                filter: dateQuery.filter,
                process: dateQuery.process,
                lrs: this.lrs,
            });
            const dateResults = await Executor.runStage(queryStage, null, this.lrs);
            console.log("dateResults", dateResults);
            const [min, max] = dateResults;

            return {
                deliveryMethods: deliveryMethods,
                reports: reports,
                courseStatus: courseStatus,
                minDate: moment(min.value).format('YYYY-MM-DD'),
                maxDate: moment(max.value).format('YYYY-MM-DD'),
                errorStatus: error,
                errorLevel: level,
            };
        };

        router.get('/courses', async (req, res, next) => {
            const searchSnippet = req.query.term;
            const snippetRegex = new RegExp(searchSnippet, 'i');

            const courseFilter = {
                $and: [
                    { "object.definition.type": "http://adlnet.gov/expapi/activities/course" },
                    {
                        "object.definition.extensions['https://dcsa-course'].delivery_method": {
                            "$in": deliveryMethodFilter(req.query.deliveryMethod),
                        }
                    },
                    {
                        "verb.id": {
                            $in: [
                                "http://adlnet.gov/expapi/verbs/completed",
                                "http://adlnet.gov/expapi/verbs/registered",
                            ]
                        }
                    },
                    {
                        "context.extensions['https://dcsa-user'].userid": { $ne: "832865" }
                    }
                ]
            }
            const courseList = (await getUniqueActivityDocs(this.lrs, courseFilter))
                .filter(c => snippetRegex.test(c.canonical.display));

            res.send(courseList);
        })

        router.get('/completedReports/:uuid/download', async (req, res, next) => {

            let connectionPool = require("./utils/connectionPool.js")
            let odm = await connectionPool.odm(this.lrs);
            let file = await odm.file.findOne({ packageID: "DCSA Report", uuid: req.params.uuid });

            res.header('Content-Type', 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet');
            res.header("Content-Disposition", `attachment; filename=${file.path}`);

            file.send(req, res);
        });
        router.get('/completedReports/:uuid/delete', async (req, res, next) => {

            let connectionPool = require("./utils/connectionPool.js")
            let odm = await connectionPool.odm(this.lrs);
            let file = await odm.file.findOne({ packageID: "DCSA Report", uuid: req.params.uuid });
            if (file)
                await file.remove();
            res.redirect("../../")
        });

        router.get('/completedReports/:uuid/cancel', async (req, res, next) => {

            let connectionPool = require("./utils/connectionPool.js")
            let odm = await connectionPool.odm(this.lrs);
            let file = await odm.file.findOne({ packageID: "DCSA Report", uuid: req.params.uuid });
            if (file) {
                file.data = "cancel";
                await file.save();
            }
            setTimeout(() => res.redirect("../../"), 2000);

        });
        router.get('/completedReports', async (req, res, next) => {

            let connectionPool = require("./utils/connectionPool.js")
            let odm = await connectionPool.odm(this.lrs);
            let files = await odm.file.find({ packageID: "DCSA Report" });

            return res._render(
                'templates/plugins/dcsaReports/downloads',
                {
                    files
                }
            );
        });

        router.get('/dcsaReports', async (req, res, next) => {
            if (req.query.error) {
                return res._render(
                    'templates/plugins/dcsaReports/manage',
                    await getHomeParams(decodeURIComponent(req.query.error), req.query.level),
                );
            }
            return res._render(
                'templates/plugins/dcsaReports/manage',
                await getHomeParams(),
            );
        });

        router.post('/dcsaReports/download', async (req, res, next) => {
            let startTimestamp = `${req.body.startDate}T00:00:00.000Z`;
            let endTimestamp = `${req.body.endDate}T23:59:59.999Z`;

            let query = {
                $and: [
                    { "object.definition.type": "http://adlnet.gov/expapi/activities/course" },
                    {
                        "timestamp": {
                            "$gte": startTimestamp,
                            "$lte": endTimestamp,
                        }
                    },
                    {
                        "object.definition.extensions['https://dcsa-course'].delivery_method": {
                            "$in": deliveryMethodFilter(req.body.deliveryMethod),
                        }
                    },
                    {
                        "verb.id": {
                            $in: [
                                "http://adlnet.gov/expapi/verbs/completed",
                                "http://adlnet.gov/expapi/verbs/registered",
                            ]
                        }
                    },
                    {
                        "context.extensions['https://dcsa-user'].userid": { $ne: "832865" }
                    }
                ]
            };

            if (req.body.selectedCourses) {
                query.$and.push({ "object.id": { $in: req.body.selectedCourses } });
            };

            const fReportName = reports.find(r => r.value === req.body.reportName) ? reports.find(r => r.value === req.body.reportName).display : 'Report';
            const fDeliveryMethod = deliveryMethods.find(r => r.value === req.body.deliveryMethod) ? deliveryMethods.find(r => r.value === req.body.deliveryMethod).display : 'Unknown';
            const fCourseStatus = courseStatus.find(r => r.value === req.body.courseStatus) ? courseStatus.find(r => r.value === req.body.courseStatus).display : 'Unknown';
            const fileName = `${fReportName}-${fDeliveryMethod} Delivered-${fCourseStatus}-${req.body.startDate} to ${req.body.endDate}.xlsx`

            let connectionPool = require("./utils/connectionPool.js")
            let odm = await connectionPool.odm(this.lrs);
            let file = new odm.file();
            file.packageID = "DCSA Report";
            file.path = req.body.reportTitle + " - " + fileName;
            file.type = "processing";
            file.uuid = require("uuid").v4();
            file.owner = this.lrs.owner;
            await file.save();

            const stream = require("stream");
            var pass = new stream.PassThrough();

            let fileProm = file.fromStream(pass);
            res.render("templates/utils/message", { messageUnsafe: "Your report is running, and will be available for download in a few minutes. <a href='../../completedReports/'> Click here to see running reports </a>" });
            const excel = require('exceljs');
            const wb = new excel.stream.xlsx.WorkbookWriter({ stream: pass });
            const sh = wb.addWorksheet('Report');
            try {
                sh.addRow([`${fDeliveryMethod} Delivered - ${fReportName} - ${fCourseStatus} - ${req.body.startDate} to ${req.body.endDate}`]).commit();

                this.lrs.stats = await utils.populateLrsStatsNow(this.lrs);

                const report = await (require(`./${req.body.reportName}`))(this.lrs, query, req.body.courseStatus, pass, file);

                if (report.length <= 0) {
                    sh.addRow(['No report data for the chosen criteria.']).commit();
                    sh.commit();
                    return wb.commit().then(async () => {
                        pass.end();
                        await fileProm;
                        file.type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
                        await file.save();
                    });
                    // return res.redirect(`../?error=No data for the chosen criteria&level=warning`);
                }

                sh.addRow(Object.keys(report[0])).commit();
                report.forEach(row => sh.addRow(Object.values(row)).commit());

                sh.commit();
                wb.commit().then(async () => {
                    pass.end();
                    await fileProm;
                    file.type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
                    await file.save();
                });

            } catch (e) {
                console.log(e);
                sh.addRow(['An error occurred while processing this report.']).commit();
                sh.commit();
                return wb.commit().then(async () => {
                    pass.end();
                    await fileProm;
                    file.type = "error";
                    await file.save();
                });
                // return res.redirect(`../?error=An error occurred&level=danger`);
            }

        });

        this.setRouter('lrs', router);
    }

    // Install will be run when the plugin is added into an LRS
    async install() {
        console.log('install me ' + this.uuid);
    }
    // Uninstall will be run when the plugin is removed from the LRS
    async uninstall() {
        console.log('uninstall me ' + this.uuid);
        // We need to make sure to call the superclass uninstall, or we'll leave behind queued events!
        super.uninstall();
    }
    // Metadata for display
    static get display() {
        return {
            title: 'DCSA Reports',
            description: 'A plugin to generate DCSA Course Reports.',
        };
    }
    // Additional metadata for display
    static get metadata() {
        return {
            author: 'Veracity Technology Consultants',
            version: '1.1.0',
            moreInfo: 'https://www.veracity.it',
        };
    }

    static get settingsForm() {
        return [

        ];
    }
}

module.exports.pluginName = 'DcsaReports';
